@@ -88,7 +88,7 @@ typedef struct
 
   GTlsCertificate       *certificate;
   gboolean               missing_requested_client_certificate;
-  GError                *certificate_error;
+  GError                *interaction_error;
   GTlsCertificate       *peer_certificate;
   GTlsCertificateFlags   peer_certificate_errors;
 
@@ -147,6 +147,8 @@ typedef struct
   GError        *write_error;
   GCancellable  *write_cancellable;
 
+  gboolean       successful_posthandshake_op;
+
   gboolean       is_system_certdb;
   gboolean       database_is_unset;
 
@@ -165,10 +167,21 @@ static gboolean do_implicit_handshake (GTlsConnectionBase  *tls,
                                        gint64               timeout,
                                        GCancellable        *cancellable,
                                        GError             **error);
+
 static gboolean finish_handshake (GTlsConnectionBase  *tls,
                                   GTask               *task,
                                   GError             **error);
 
+static void g_tls_connection_base_handshake_async (GTlsConnection      *conn,
+                                                   int                  io_priority,
+                                                   GCancellable        *cancellable,
+                                                   GAsyncReadyCallback  callback,
+                                                   gpointer             user_data);
+
+static gboolean g_tls_connection_base_handshake (GTlsConnection   *conn,
+                                                 GCancellable     *cancellable,
+                                                 GError          **error);
+
 G_DEFINE_ABSTRACT_TYPE_WITH_CODE (GTlsConnectionBase, g_tls_connection_base, G_TYPE_TLS_CONNECTION,
                                   G_ADD_PRIVATE (GTlsConnectionBase);
                                   G_IMPLEMENT_INTERFACE (G_TYPE_DATAGRAM_BASED,
@@ -238,7 +251,7 @@ g_tls_connection_base_finalize (GObject *object)
 
   g_clear_object (&priv->database);
   g_clear_object (&priv->certificate);
-  g_clear_error (&priv->certificate_error);
+  g_clear_error (&priv->interaction_error);
   g_clear_object (&priv->peer_certificate);
 
   g_mutex_clear (&priv->verify_certificate_mutex);
@@ -357,8 +370,7 @@ g_tls_connection_base_set_property (GObject      *object,
   switch (prop_id)
     {
     case PROP_BASE_IO_STREAM:
-      g_assert (g_value_get_object (value) == NULL ||
-                priv->base_socket == NULL);
+      g_assert (!g_value_get_object (value) || !priv->base_socket);
 
       if (priv->base_io_stream)
         {
@@ -388,8 +400,7 @@ g_tls_connection_base_set_property (GObject      *object,
       break;
 
     case PROP_BASE_SOCKET:
-      g_assert (g_value_get_object (value) == NULL ||
-                priv->base_io_stream == NULL);
+      g_assert (!g_value_get_object (value) || !priv->base_io_stream);
 
       g_clear_object (&priv->base_socket);
       priv->base_socket = g_value_dup_object (value);
@@ -724,6 +735,7 @@ g_tls_connection_base_real_pop_io (GTlsConnectionBase  *tls,
       else
         g_clear_error (&priv->read_error);
     }
+
   if (direction & G_IO_OUT)
     {
       priv->write_cancellable = NULL;
@@ -744,13 +756,43 @@ g_tls_connection_base_real_pop_io (GTlsConnectionBase  *tls,
       g_propagate_error (error, my_error);
       return G_TLS_CONNECTION_BASE_WOULD_BLOCK;
     }
-  else if (g_error_matches (my_error, G_IO_ERROR, G_IO_ERROR_TIMED_OUT))
+
+  if (g_error_matches (my_error, G_IO_ERROR, G_IO_ERROR_TIMED_OUT))
     {
       g_propagate_error (error, my_error);
       return G_TLS_CONNECTION_BASE_TIMED_OUT;
     }
+
+  if (priv->missing_requested_client_certificate &&
+      !priv->successful_posthandshake_op)
+    {
+      g_assert (G_IS_TLS_CLIENT_CONNECTION (tls));
+
+      /* Probably the server requires a client certificate, but we failed to
+       * provide one. With TLS 1.3 the server is no longer able to tell us
+       * this, so we just have to guess. If there is an error from the TLS
+       * interaction (request for user certificate), we provide that. Otherwise,
+       * guess that G_TLS_ERROR_CERTIFICATE_REQUIRED is probably appropriate.
+       * This could be wrong, but only applies to the small minority of
+       * connections where a client cert is requested but not provided, and then
+       * then only if the client has never successfully read or written.
+       */
+      if (priv->interaction_error)
+        {
+          g_propagate_error (error, priv->interaction_error);
+          priv->interaction_error = NULL;
+        }
+      else
+        {
+          g_clear_error (error);
+          g_set_error_literal (error, G_TLS_ERROR, G_TLS_ERROR_CERTIFICATE_REQUIRED,
+                               _("Server required TLS certificate"));
+        }
+    }
   else if (my_error)
-    g_propagate_error (error, my_error);
+    {
+      g_propagate_error (error, my_error);
+    }
 
   return G_TLS_CONNECTION_BASE_ERROR;
 }
@@ -832,20 +874,6 @@ typedef struct {
   gboolean            op_waiting;
 } GTlsConnectionBaseSource;
 
-static gboolean
-tls_source_prepare (GSource *source,
-                    gint    *timeout)
-{
-  *timeout = -1;
-  return FALSE;
-}
-
-static gboolean
-tls_source_check (GSource *source)
-{
-  return FALSE;
-}
-
 /* Use a custom dummy callback instead of g_source_set_dummy_callback(), as that
  * uses a GClosure and is slow. (The GClosure is necessary to deal with any
  * function prototype.) */
@@ -995,8 +1023,8 @@ g_tls_connection_tls_source_dtls_closure_callback (GObject      *stream,
 
 static GSourceFuncs tls_source_funcs =
 {
-  tls_source_prepare,
-  tls_source_check,
+  NULL,
+  NULL,
   tls_source_dispatch,
   tls_source_finalize,
   (GSourceFunc)g_tls_connection_tls_source_closure_callback,
@@ -1005,8 +1033,8 @@ static GSourceFuncs tls_source_funcs =
 
 static GSourceFuncs dtls_source_funcs =
 {
-  tls_source_prepare,
-  tls_source_check,
+  NULL,
+  NULL,
   tls_source_dispatch,
   tls_source_finalize,
   (GSourceFunc)g_tls_connection_tls_source_dtls_closure_callback,
@@ -1038,9 +1066,9 @@ g_tls_connection_base_create_source (GTlsConnectionBase  *tls,
   tls_source->condition = condition;
   if (g_tls_connection_base_is_dtls (tls))
     tls_source->base = G_OBJECT (tls);
-  else if (priv->tls_istream != NULL && condition & G_IO_IN)
+  else if (priv->tls_istream && condition & G_IO_IN)
     tls_source->base = G_OBJECT (priv->tls_istream);
-  else if (priv->tls_ostream != NULL && condition & G_IO_OUT)
+  else if (priv->tls_ostream && condition & G_IO_OUT)
     tls_source->base = G_OBJECT (priv->tls_ostream);
   else
     g_assert_not_reached ();
@@ -1160,7 +1188,7 @@ verify_peer_certificate (GTlsConnectionBase *tls,
   errors = 0;
 
   database = g_tls_connection_get_database (G_TLS_CONNECTION (tls));
-  if (database == NULL)
+  if (!database)
     {
       errors |= G_TLS_CERTIFICATE_UNKNOWN_CA;
       errors |= g_tls_certificate_verify (peer_certificate, peer_identity, NULL);
@@ -1233,7 +1261,7 @@ accept_or_reject_peer_certificate (gpointer user_data)
 
   update_peer_certificate_and_compute_errors (tls);
 
-  if (G_IS_TLS_CLIENT_CONNECTION (tls) && priv->peer_certificate != NULL)
+  if (G_IS_TLS_CLIENT_CONNECTION (tls) && priv->peer_certificate)
     {
       GTlsCertificateFlags validation_flags;
 
@@ -1328,7 +1356,7 @@ handshake_thread (GTask        *task,
   gint64 timeout;
 
   /* A timeout, in microseconds, must be provided as a gint64* task_data. */
-  g_assert (task_data != NULL);
+  g_assert (task_data);
   start_time = g_get_monotonic_time ();
   timeout = *((gint64 *)task_data);
 
@@ -1348,6 +1376,14 @@ handshake_thread (GTask        *task,
     {
       GTlsConnectionBaseStatus status;
 
+      if (priv->rehandshake_mode != G_TLS_REHANDSHAKE_UNSAFELY &&
+          tls_class->handshake_thread_safe_renegotiation_status (tls) != G_TLS_SAFE_RENEGOTIATION_SUPPORTED_BY_PEER)
+        {
+          g_task_return_new_error (task, G_TLS_ERROR, G_TLS_ERROR_MISC,
+                                   _("Peer does not support safe renegotiation"));
+          return;
+        }
+
       /* Adjust the timeout for the next operation in the sequence. */
       if (timeout > 0)
         {
@@ -1356,7 +1392,7 @@ handshake_thread (GTask        *task,
             timeout = 1;
         }
 
-      status = tls_class->request_rehandshake (tls, timeout, cancellable, &error);
+      status = tls_class->handshake_thread_request_rehandshake (tls, timeout, cancellable, &error);
       if (status != G_TLS_CONNECTION_BASE_OK)
         {
           g_task_return_error (task, error);
@@ -1376,21 +1412,6 @@ handshake_thread (GTask        *task,
   tls_class->handshake_thread_handshake (tls, timeout, cancellable, &error);
   priv->need_handshake = FALSE;
 
-  if (priv->missing_requested_client_certificate)
-    {
-      g_clear_error (&error);
-      if (priv->certificate_error)
-        {
-          error = priv->certificate_error;
-          priv->certificate_error = NULL;
-        }
-      else
-        {
-          g_set_error_literal (&error, G_TLS_ERROR, G_TLS_ERROR_CERTIFICATE_REQUIRED,
-                               _("Server required TLS certificate"));
-        }
-    }
-
   if (error)
     {
       g_task_return_error (task, error);
@@ -1409,6 +1430,11 @@ sync_handshake_thread_completed (GObject      *object,
 {
   GTlsConnectionBase *tls = G_TLS_CONNECTION_BASE (object);
   GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
+  gpointer source_tag;
+
+  source_tag = g_task_get_source_tag (G_TASK (result));
+  g_assert (source_tag == do_implicit_handshake || source_tag == g_tls_connection_base_handshake);
+  g_assert (g_task_is_valid (result, object));
 
   g_assert (g_main_context_is_owner (priv->handshake_context));
 
@@ -1509,7 +1535,7 @@ g_tls_connection_base_handshake (GTlsConnection   *conn,
   gint64 *timeout = NULL;
   GError *my_error = NULL;
 
-  g_assert (priv->handshake_context == NULL);
+  g_assert (!priv->handshake_context);
   priv->handshake_context = g_main_context_new ();
 
   g_main_context_push_thread_default (priv->handshake_context);
@@ -1519,6 +1545,7 @@ g_tls_connection_base_handshake (GTlsConnection   *conn,
 
   task = g_task_new (conn, cancellable, sync_handshake_thread_completed, NULL);
   g_task_set_source_tag (task, g_tls_connection_base_handshake);
+  g_task_set_name (task, "[glib-networking] g_tls_connection_base_handshake");
   g_task_set_return_on_cancel (task, TRUE);
 
   timeout = g_new0 (gint64, 1);
@@ -1568,6 +1595,9 @@ handshake_thread_completed (GObject      *object,
   GError *error = NULL;
   gboolean need_finish_handshake, success;
 
+  g_assert (g_task_is_valid (result, object));
+  g_assert (g_task_get_source_tag (G_TASK (result)) == g_tls_connection_base_handshake_async);
+
   g_mutex_lock (&priv->op_mutex);
   if (priv->need_finish_handshake)
     {
@@ -1640,10 +1670,12 @@ g_tls_connection_base_handshake_async (GTlsConnection      *conn,
 
   caller_task = g_task_new (conn, cancellable, callback, user_data);
   g_task_set_source_tag (caller_task, g_tls_connection_base_handshake_async);
+  g_task_set_name (caller_task, "[glib-networking] g_tls_connection_base_handshake_async (caller task)");
   g_task_set_priority (caller_task, io_priority);
 
   thread_task = g_task_new (conn, cancellable, handshake_thread_completed, caller_task);
   g_task_set_source_tag (thread_task, g_tls_connection_base_handshake_async);
+  g_task_set_name (caller_task, "[glib-networking] g_tls_connection_base_handshake_async (thread task)");
   g_task_set_priority (thread_task, io_priority);
 
   timeout = g_new0 (gint64, 1);
@@ -1660,6 +1692,7 @@ g_tls_connection_base_handshake_finish (GTlsConnection  *conn,
                                         GError         **error)
 {
   g_return_val_if_fail (g_task_is_valid (result, conn), FALSE);
+  g_return_val_if_fail (g_task_get_source_tag (G_TASK (result)) == g_tls_connection_base_handshake_async, FALSE);
 
   return g_task_propagate_boolean (G_TASK (result), error);
 }
@@ -1696,7 +1729,7 @@ do_implicit_handshake (GTlsConnectionBase  *tls,
 
   /* We have op_mutex */
 
-  g_assert (priv->handshake_context == NULL);
+  g_assert (!priv->handshake_context);
   if (timeout != 0)
     {
       priv->handshake_context = g_main_context_new ();
@@ -1707,11 +1740,12 @@ do_implicit_handshake (GTlsConnectionBase  *tls,
       priv->handshake_context = g_main_context_ref_thread_default ();
     }
 
-  g_assert (priv->implicit_handshake == NULL);
+  g_assert (!priv->implicit_handshake);
   priv->implicit_handshake = g_task_new (tls, cancellable,
                                         timeout ? sync_handshake_thread_completed : NULL,
                                         NULL);
   g_task_set_source_tag (priv->implicit_handshake, do_implicit_handshake);
+  g_task_set_name (priv->implicit_handshake, "[glib-networking] do_implicit_handshake");
 
   thread_timeout = g_new0 (gint64, 1);
   g_task_set_task_data (priv->implicit_handshake,
@@ -1813,9 +1847,12 @@ g_tls_connection_base_read (GTlsConnectionBase  *tls,
   while (status == G_TLS_CONNECTION_BASE_REHANDSHAKE);
 
   if (status == G_TLS_CONNECTION_BASE_OK)
-    return nread;
-  else
-    return -1;
+    {
+      priv->successful_posthandshake_op = TRUE;
+      return nread;
+    }
+
+  return -1;
 }
 
 static gssize
@@ -1867,7 +1904,11 @@ g_tls_connection_base_read_message (GTlsConnectionBase  *tls,
   } while (status == G_TLS_CONNECTION_BASE_REHANDSHAKE);
 
   if (status == G_TLS_CONNECTION_BASE_OK)
-    return nread;
+    {
+      priv->successful_posthandshake_op = TRUE;
+      return nread;
+    }
+
   return -1;
 }
 
@@ -1880,12 +1921,11 @@ g_tls_connection_base_receive_messages (GDatagramBased  *datagram_based,
                                         GCancellable    *cancellable,
                                         GError         **error)
 {
-  GTlsConnectionBase *tls;
+  GTlsConnectionBase *tls = G_TLS_CONNECTION_BASE (datagram_based);
+  GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
   guint i;
   GError *child_error = NULL;
 
-  tls = G_TLS_CONNECTION_BASE (datagram_based);
-
   if (flags != G_SOCKET_MSG_NONE)
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
@@ -1893,7 +1933,7 @@ g_tls_connection_base_receive_messages (GDatagramBased  *datagram_based,
       return -1;
     }
 
-  for (i = 0; i < num_messages && child_error == NULL; i++)
+  for (i = 0; i < num_messages && !child_error; i++)
     {
       GInputMessage *message = &messages[i];
       gssize n_bytes_read;
@@ -1905,10 +1945,10 @@ g_tls_connection_base_receive_messages (GDatagramBased  *datagram_based,
                                                          cancellable,
                                                          &child_error);
 
-      if (message->address != NULL)
+      if (message->address)
         *message->address = NULL;
       message->flags = G_SOCKET_MSG_NONE;
-      if (message->control_messages != NULL)
+      if (message->control_messages)
         *message->control_messages = NULL;
       message->num_control_messages = 0;
 
@@ -1939,12 +1979,13 @@ g_tls_connection_base_receive_messages (GDatagramBased  *datagram_based,
         }
     }
 
-  if (child_error != NULL)
+  if (child_error)
     {
       g_propagate_error (error, child_error);
       return -1;
     }
 
+  priv->successful_posthandshake_op = TRUE;
   return i;
 }
 
@@ -1956,6 +1997,7 @@ g_tls_connection_base_write (GTlsConnectionBase  *tls,
                              GCancellable        *cancellable,
                              GError             **error)
 {
+  GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
   GTlsConnectionBaseStatus status;
   gssize nwrote;
 
@@ -1973,9 +2015,12 @@ g_tls_connection_base_write (GTlsConnectionBase  *tls,
   while (status == G_TLS_CONNECTION_BASE_REHANDSHAKE);
 
   if (status == G_TLS_CONNECTION_BASE_OK)
-    return nwrote;
-  else
-    return -1;
+    {
+      priv->successful_posthandshake_op = TRUE;
+      return nwrote;
+    }
+
+  return -1;
 }
 
 static gssize
@@ -1986,6 +2031,7 @@ g_tls_connection_base_write_message (GTlsConnectionBase  *tls,
                                      GCancellable        *cancellable,
                                      GError             **error)
 {
+  GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
   GTlsConnectionBaseStatus status;
   gssize nwrote;
 
@@ -2002,7 +2048,11 @@ g_tls_connection_base_write_message (GTlsConnectionBase  *tls,
   } while (status == G_TLS_CONNECTION_BASE_REHANDSHAKE);
 
   if (status == G_TLS_CONNECTION_BASE_OK)
-    return nwrote;
+    {
+      priv->successful_posthandshake_op = TRUE;
+      return nwrote;
+    }
+
   return -1;
 }
 
@@ -2015,12 +2065,11 @@ g_tls_connection_base_send_messages (GDatagramBased  *datagram_based,
                                      GCancellable    *cancellable,
                                      GError         **error)
 {
-  GTlsConnectionBase *tls;
+  GTlsConnectionBase *tls = G_TLS_CONNECTION_BASE (datagram_based);
+  GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
   guint i;
   GError *child_error = NULL;
 
-  tls = G_TLS_CONNECTION_BASE (datagram_based);
-
   if (flags != G_SOCKET_MSG_NONE)
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
@@ -2028,7 +2077,7 @@ g_tls_connection_base_send_messages (GDatagramBased  *datagram_based,
       return -1;
     }
 
-  for (i = 0; i < num_messages && child_error == NULL; i++)
+  for (i = 0; i < num_messages && !child_error; i++)
     {
       GOutputMessage *message = &messages[i];
       gssize n_bytes_sent;
@@ -2062,12 +2111,13 @@ g_tls_connection_base_send_messages (GDatagramBased  *datagram_based,
         }
     }
 
-  if (child_error != NULL)
+  if (child_error)
     {
       g_propagate_error (error, child_error);
       return -1;
     }
 
+  priv->successful_posthandshake_op = TRUE;
   return i;
 }
 
@@ -2138,7 +2188,7 @@ g_tls_connection_base_close_internal (GIOStream      *stream,
   /* Close the underlying streams. Do this even if the close_fn() call failed,
    * as the parent GIOStream will have set its internal closed flag and hence
    * this implementation will never be called again. */
-  if (priv->base_io_stream != NULL)
+  if (priv->base_io_stream)
     {
       if (direction == G_TLS_DIRECTION_BOTH)
         success = g_io_stream_close (priv->base_io_stream,
@@ -2246,6 +2296,7 @@ g_tls_connection_base_close_internal_async (GIOStream           *stream,
 
   task = g_task_new (stream, cancellable, callback, user_data);
   g_task_set_source_tag (task, g_tls_connection_base_close_internal_async);
+  g_task_set_name (task, "[glib-networking] g_tls_connection_base_close_internal_async");
   g_task_set_priority (task, io_priority);
   g_task_set_task_data (task, GINT_TO_POINTER (direction), NULL);
   g_task_run_in_thread (task, close_thread);
@@ -2270,6 +2321,7 @@ g_tls_connection_base_close_finish (GIOStream           *stream,
                                     GError             **error)
 {
   g_return_val_if_fail (g_task_is_valid (result, stream), FALSE);
+  g_return_val_if_fail (g_task_get_source_tag (G_TASK (result)) == g_tls_connection_base_close_internal_async, FALSE);
 
   return g_task_propagate_boolean (G_TASK (result), error);
 }
@@ -2301,6 +2353,7 @@ g_tls_connection_base_dtls_shutdown_finish (GDtlsConnection  *conn,
                                             GError          **error)
 {
   g_return_val_if_fail (g_task_is_valid (result, conn), FALSE);
+  g_return_val_if_fail (g_task_get_source_tag (G_TASK (result)) == g_tls_connection_base_close_internal_async, FALSE);
 
   return g_task_propagate_boolean (G_TASK (result), error);
 }
@@ -2367,19 +2420,10 @@ g_tls_connection_base_set_missing_requested_client_certificate (GTlsConnectionBa
   GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
 
   /* FIXME: Assert this is only used on the handshake thread. */
-  /* FIXME: Assert it's not already requested? Probably. */
 
   priv->missing_requested_client_certificate = TRUE;
 }
 
-GError **
-g_tls_connection_base_get_certificate_error (GTlsConnectionBase *tls)
-{
-  GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
-
-  return &priv->certificate_error;
-}
-
 GError **
 g_tls_connection_base_get_read_error (GTlsConnectionBase *tls)
 {
@@ -2445,8 +2489,7 @@ g_tls_connection_base_ever_handshaked (GTlsConnectionBase *tls)
 }
 
 gboolean
-g_tls_connection_base_request_certificate (GTlsConnectionBase  *tls,
-                                           GError             **error)
+g_tls_connection_base_request_certificate (GTlsConnectionBase *tls)
 {
   GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
   GTlsInteractionResult res = G_TLS_INTERACTION_UNHANDLED;
@@ -2457,24 +2500,25 @@ g_tls_connection_base_request_certificate (GTlsConnectionBase  *tls,
 
   conn = G_TLS_CONNECTION (tls);
 
+  g_clear_error (&priv->interaction_error);
+
   interaction = g_tls_connection_get_interaction (conn);
   if (!interaction)
     return FALSE;
 
   res = g_tls_interaction_invoke_request_certificate (interaction, conn, 0,
-                                                      priv->read_cancellable, error);
+                                                      priv->read_cancellable,
+                                                      &priv->interaction_error);
   return res != G_TLS_INTERACTION_FAILED;
 }
 
 void
-g_tls_connection_base_buffer_application_data (GTlsConnectionBase *tls,
-                                               guint8             *data,
-                                               gsize               length)
+g_tls_connection_base_handshake_thread_buffer_application_data (GTlsConnectionBase *tls,
+                                                                guint8             *data,
+                                                                gsize               length)
 {
   GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
 
-  /* FIXME: Called from handshake thread, needs a mutex! */
-
   if (!priv->app_data_buf)
     priv->app_data_buf = g_byte_array_new ();
 
