@@ -89,9 +89,7 @@ typedef struct {
   gboolean expect_server_error;
   GError *server_error;
   gboolean server_running;
-#if GLIB_CHECK_VERSION(2, 60, 0)
   const gchar * const *server_protocols;
-#endif
 
   char buf[128];
   gssize nread, nwrote;
@@ -400,13 +398,11 @@ on_incoming_connection (GSocket       *socket,
   if (test->database)
     g_dtls_connection_set_database (G_DTLS_CONNECTION (test->server_connection), test->database);
 
-#if GLIB_CHECK_VERSION(2, 60, 0)
   if (test->server_protocols)
     {
       g_dtls_connection_set_advertised_protocols (G_DTLS_CONNECTION (test->server_connection),
                                                   test->server_protocols);
     }
-#endif
 
   if (test->test_data->server_should_disappear)
     {
@@ -743,7 +739,6 @@ test_alpn (TestConnection *test,
            const char * const *server_protocols,
            const char *negotiated_protocol)
 {
-#if GLIB_CHECK_VERSION(2, 60, 0)
   GDatagramBased *connection;
   GError *error = NULL;
 
@@ -775,9 +770,6 @@ test_alpn (TestConnection *test,
 
   g_assert_cmpstr (g_dtls_connection_get_negotiated_protocol (G_DTLS_CONNECTION (test->server_connection)), ==, negotiated_protocol);
   g_assert_cmpstr (g_dtls_connection_get_negotiated_protocol (G_DTLS_CONNECTION (test->client_connection)), ==, negotiated_protocol);
-#else
-  g_test_skip ("no support for ALPN in this GLib version");
-#endif
 }
 
 static void
@@ -874,37 +866,37 @@ main (int   argc,
   g_setenv ("GIO_USE_TLS", BACKEND, TRUE);
   g_assert (g_ascii_strcasecmp (G_OBJECT_TYPE_NAME (g_tls_backend_get_default ()), "GTlsBackend" BACKEND) == 0);
 
-  g_test_add ("/dtls/connection/basic/blocking", TestConnection, &blocking,
+  g_test_add ("/dtls/" BACKEND "/connection/basic/blocking", TestConnection, &blocking,
               setup_connection, test_basic_connection, teardown_connection);
-  g_test_add ("/dtls/connection/basic/timeout", TestConnection, &server_timeout,
+  g_test_add ("/dtls/" BACKEND "/connection/basic/timeout", TestConnection, &server_timeout,
               setup_connection, test_basic_connection, teardown_connection);
-  g_test_add ("/dtls/connection/basic/nonblocking",
+  g_test_add ("/dtls/" BACKEND "/connection/basic/nonblocking",
               TestConnection, &nonblocking,
               setup_connection, test_basic_connection, teardown_connection);
 
-  g_test_add ("/dtls/connection/threaded/blocking", TestConnection, &blocking,
+  g_test_add ("/dtls/" BACKEND "/connection/threaded/blocking", TestConnection, &blocking,
               setup_connection, test_threaded_connection, teardown_connection);
-  g_test_add ("/dtls/connection/threaded/timeout",
+  g_test_add ("/dtls/" BACKEND "/connection/threaded/timeout",
               TestConnection, &server_timeout,
               setup_connection, test_threaded_connection, teardown_connection);
-  g_test_add ("/dtls/connection/threaded/nonblocking",
+  g_test_add ("/dtls/" BACKEND "/connection/threaded/nonblocking",
               TestConnection, &nonblocking,
               setup_connection, test_threaded_connection, teardown_connection);
 
-  g_test_add ("/dtls/connection/timeouts/read", TestConnection, &client_timeout,
+  g_test_add ("/dtls/" BACKEND "/connection/timeouts/read", TestConnection, &client_timeout,
               setup_connection, test_connection_timeouts_read,
               teardown_connection);
 
-  g_test_add ("/dtls/connection/alpn/match", TestConnection, &blocking,
+  g_test_add ("/dtls/" BACKEND "/connection/alpn/match", TestConnection, &blocking,
               setup_connection, test_alpn_match,
               teardown_connection);
-  g_test_add ("/dtls/connection/alpn/no-match", TestConnection, &blocking,
+  g_test_add ("/dtls/" BACKEND "/connection/alpn/no-match", TestConnection, &blocking,
               setup_connection, test_alpn_no_match,
               teardown_connection);
-  g_test_add ("/dtls/connection/alpn/client-only", TestConnection, &blocking,
+  g_test_add ("/dtls/" BACKEND "/connection/alpn/client-only", TestConnection, &blocking,
               setup_connection, test_alpn_client_only,
               teardown_connection);
-  g_test_add ("/dtls/connection/alpn/server-only", TestConnection, &blocking,
+  g_test_add ("/dtls/" BACKEND "/connection/alpn/server-only", TestConnection, &blocking,
               setup_connection, test_alpn_server_only,
               teardown_connection);
 
